import os
from typing import List, Tuple
import numpy as np
from enum import Enum, auto

from blueness import module
from blue_options import string
from blue_objects import file, path, objects
from blue_objects.metadata import get_from_object
from blue_objects.storage import instance as storage

from roofai import NAME
from roofai.dataset.ingest.CamVid import CLASSES as CAMVID_CLASSES
from roofai.semseg import (
    chip_width as semseg_chip_width,
    chip_height as semseg_chip_height,
)
from roofai.dataset.sagemaker import (
    chip_width as sagesemseg_chip_width,
    chip_height as sagesemseg_chip_height,
)
from roofai.semseg.utils import visualize
from roofai.logger import logger

NAME = module.name(__file__, NAME)


class DatasetTarget(Enum):
    TORCH = auto()
    SAGEMAKER = auto()

    @property
    def chip_height(self) -> int:
        if self == DatasetTarget.TORCH:
            return semseg_chip_height

        assert self == DatasetTarget.SAGEMAKER
        return sagesemseg_chip_height

    @property
    def chip_width(self) -> int:
        if self == DatasetTarget.TORCH:
            return semseg_chip_width

        assert self == DatasetTarget.SAGEMAKER
        return sagesemseg_chip_width


class MatrixKind(Enum):
    IMAGE = auto()
    MASK = auto()

    def filename_and_extension(
        self,
        dataset_kind,
        subset,
        record_id,
    ):
        return "{}/{}.{}".format(
            self.subset_path(dataset_kind, subset),
            record_id,
            dataset_kind.file_extension(self),
        )

    def subset_path(
        self,
        dataset_kind,
        subset,
    ) -> str:
        sagemaker_val_adapter = lambda subset: (
            "validation" if subset == "val" else subset
        )

        return (
            (subset if self == MatrixKind.IMAGE else f"{subset}annot")
            if dataset_kind == DatasetKind.CAMVID
            else (
                (
                    sagemaker_val_adapter(subset)
                    if self == MatrixKind.IMAGE
                    else f"{sagemaker_val_adapter(subset)}_annotation"
                )
                if dataset_kind == DatasetKind.SAGEMAKER
                else (
                    f"{subset}/image" if self == MatrixKind.IMAGE else f"{subset}/label"
                )
            )
        )


class DatasetKind(Enum):
    AIRS = auto()
    CAMVID = auto()
    SAGEMAKER = auto()
    DISTRIBUTED = auto()

    def file_extension(self, kind: MatrixKind) -> str:
        if self in [DatasetKind.AIRS, DatasetKind.DISTRIBUTED]:
            return "tif"

        if self == DatasetKind.CAMVID:
            return "png"

        assert self == DatasetKind.SAGEMAKER
        if kind == MatrixKind.MASK:
            return "png"

        assert kind == MatrixKind.IMAGE
        return "jpg"

    @property
    def prefix_path(self) -> str:
        return "SegNet-Tutorial/CamVid/" if self == DatasetKind.CAMVID else ""


class RoofAIDataset:
    SUBSETS = "test,train,val".split(",")

    def __init__(self, dataset_path, kind=None):
        self.path = dataset_path
        self.object_name = path.name(self.path)

        _, self.metadata = file.load_yaml(
            os.path.join(self.path, "metadata.yaml"),
            ignore_error=True,
        )

        self.source = self.metadata.get("source", "AIRS")
        self.kind = DatasetKind[
            self.metadata.get(
                "kind",
                (
                    self.source
                    if kind is None
                    else kind.name if isinstance(kind, DatasetKind) else kind
                ),
            ).upper()
        ]

        if self.kind == DatasetKind.DISTRIBUTED:
            if self.metadata["datacube_id"]:
                datacube_id = self.metadata["datacube_id"][0]
                self.classes = get_from_object(
                    object_name=datacube_id,
                    key="rasterize",
                    default={},
                    download=True,
                ).get("list_of_classes", [])
            else:
                self.classes = []
        else:
            self.classes = self.metadata.get(
                "classes",
                "other,roof".split(",") if self.source == "AIRS" else CAMVID_CLASSES,
            )

        self.subsets = {
            subset: self.list_of_record_id(subset) for subset in self.SUBSETS
        }

        logger.info(self.one_liner)

    def create(self, log: bool = False):
        for subset in self.subsets:
            for matrix_kind in list(MatrixKind):
                path.create(
                    self.subset_path(subset, matrix_kind),
                    log=log,
                )
        return self

    def subset_path(
        self,
        subset: str,
        matrix_kind: MatrixKind,
    ) -> str:
        return os.path.join(
            self.dataset_path,
            matrix_kind.subset_path(self.kind, subset),
        )

    def list_of_record_id(
        self,
        subset: str,
        matrix_kind: MatrixKind = MatrixKind.IMAGE,
    ) -> List[str]:
        if self.kind == DatasetKind.DISTRIBUTED:
            return self.metadata.get("datacube_id", []) if subset == "train" else []

        return sorted(
            [
                file.name(filename)
                for filename in file.list_of(
                    os.path.join(
                        self.subset_path(subset, matrix_kind),
                        f"*.{self.kind.file_extension(matrix_kind)}",
                    )
                )
            ]
        )

    @property
    def one_liner(self) -> str:
        return "{}[kind:{},source:{}]({}): {} subset(s): {} - {} class(es): {}".format(
            self.__class__.__name__,
            self.kind,
            self.source,
            self.object_name,
            len(self.subsets),
            " + ".join(
                [
                    "{:,d} X {}".format(len(ids), subset)
                    for subset, ids in self.subsets.items()
                ]
            ),
            len(self.classes),
            ", ".join(self.classes),
        )

    @property
    def dataset_path(self) -> str:
        return os.path.join(self.path, self.kind.prefix_path)

    def get_filename(
        self,
        subset: str,
        record_id: str,
        matrix_kind: MatrixKind = MatrixKind.IMAGE,
        log: bool = False,
    ) -> str:
        assert isinstance(matrix_kind, MatrixKind)

        if self.kind == DatasetKind.DISTRIBUTED:
            record_metadata = get_from_object(
                object_name=record_id,
                key="rasterize",
                default={},
                download=True,
            )

            assert (
                "reference_filename" in record_metadata
            ), 'no reference filename in the record, run "@datacube label".'
            filename = record_metadata["reference_filename"]
            if matrix_kind == MatrixKind.MASK:
                filename = file.add_suffix(filename, "label")

            assert storage.download_file(f"bolt/{record_id}/{filename}", "object")

            filename = objects.path_of(filename, record_id)
        else:
            filename = os.path.join(
                self.dataset_path,
                matrix_kind.filename_and_extension(self.kind, subset, record_id),
            )

        if log:
            logger.info(
                "{}[{}].get_filename({},{},{}): {}".format(
                    self.__class__.__name__,
                    self.object_name,
                    subset,
                    record_id,
                    matrix_kind,
                    filename,
                )
            )

        return filename

    def get_matrix(
        self,
        subset: str,
        record_id: str,
        kind: MatrixKind = MatrixKind.IMAGE,
        log: bool = False,
    ):
        success, matrix = file.load_image(
            self.get_filename(subset, record_id, kind, log)
        )
        assert success

        unique_value = np.array([])
        if kind == MatrixKind.MASK:
            matrix = matrix[:, :, 0]
            unique_value = np.unique(matrix)

        if log:
            logger.info(
                "{}[{}].get_matrix({},{},{}): {}{}".format(
                    self.__class__.__name__,
                    self.object_name,
                    subset,
                    record_id,
                    kind,
                    string.pretty_shape_of_matrix(matrix),
                    (
                        ": {} unique value(s): {}".format(
                            len(unique_value),
                            unique_value,
                        )
                        if kind == MatrixKind.MASK
                        else ""
                    ),
                )
            )

        return matrix

    def shape(
        self,
        subset: str = "train",
        index: int = 0,
        kind: MatrixKind = MatrixKind.IMAGE,
        log: bool = False,
    ):
        return self.get_matrix(
            subset=subset,
            record_id=self.subsets[subset][index],
            kind=kind,
            log=log,
        ).shape

    def visualize(
        self,
        subset: str,
        index: Tuple[int, str],
        filename: str = "auto",
        in_notebook: bool = False,
        description: List[str] = [],
        log: bool = False,
    ):
        if index >= len(self.subsets[subset]):
            logger.warning(
                "{}.visualize: item ignored, index={} > len(subset[{}]) = {}.".format(
                    NAME,
                    index,
                    subset,
                    len(self.subsets[subset]),
                )
            )
            return

        record_id = self.subsets[subset][index] if isinstance(index, int) else index
        logger.info("record_id: {}".format(record_id))

        image = self.get_matrix(
            subset,
            record_id,
            MatrixKind.IMAGE,
            log=log,
        )

        mask = self.get_matrix(
            subset,
            record_id,
            MatrixKind.MASK,
            log=log,
        )

        visualize(
            {
                "image": image,
                "mask": mask,
            },
            filename=(
                os.path.join(
                    self.path,
                    f"_review/{record_id}.png",
                )
                if filename == "auto"
                else filename
            ),
            in_notebook=in_notebook,
            description=[
                self.object_name,
                f"{subset}/#{index}:{record_id}",
                f"kind: {self.kind.name.lower()}",
            ]
            + description,
        )



import os

from blue_options import MARQUEE as default_MARQUEE
from blue_objects import file, README

from roofai import NAME, VERSION, ICON, REPO_NAME
from roofai.dataset.README import items as dataset_items

features = {
    "datasets": {
        "description": "Semantic Segmentation Datasets",
        "icon": ICON,
        "thumbnail": "https://github.com/kamangir/assets/blob/main/roofAI/AIRS-cache-v45--review-index-2.png?raw=true",
        "url": "https://github.com/kamangir/roofai/blob/main/roofai/dataset",
    },
    "semseg": {
        "description": "A Semantic Segmenter based on [segmentation_models.pytorch](<https://github.com/qubvel/segmentation_models.pytorch/blob/master/examples/cars%20segmentation%20(camvid).ipynb>).",
        "icon": ICON,
        "thumbnail": "https://github.com/kamangir/roofAI/raw/main/assets/predict-00247.png",
        "url": "https://github.com/kamangir/roofai/blob/main/roofai/semseg",
    },
    "template": {
        "description": "",
        "icon": ICON,
        "thumbnail": default_MARQUEE,
        "url": "",
    },
}


items = [
    "{}[`{}`]({}) [![image]({})]({}) {}".format(
        details["icon"],
        feature,
        details["url"],
        details["thumbnail"],
        details["url"],
        details["description"],
    )
    for feature, details in features.items()
    if feature != "template"
]


def build():
    return all(
        [
            README.build(
                items=items,
                path=os.path.join(file.path(__file__), ".."),
                ICON=ICON,
                NAME=NAME,
                VERSION=VERSION,
                REPO_NAME=REPO_NAME,
            ),
            README.build(
                items=dataset_items,
                cols=len(dataset_items),
                path=os.path.join(file.path(__file__), "dataset"),
                ICON=ICON,
                NAME=NAME,
                VERSION=VERSION,
                REPO_NAME=REPO_NAME,
            ),
        ]
    )



list_of_datasets = {
    "AIRS": {
        "description": "Aerial Imagery for Roof Segmentation from [kaggle](https://www.kaggle.com/datasets/atilol/aerialimageryforroofsegmentation).",
        "thumbnail": "https://github.com/kamangir/assets/blob/main/roofAI/AIRS-cache-v45--review-index-2.png?raw=true",
    },
    "CamVid": {
        "description": "From [SegNet-Tutorial](https://github.com/alexgkendall/SegNet-Tutorial)",
        "thumbnail": "https://github.com/kamangir/assets/blob/main/roofAI/0001TP_008850.png?raw=true",
    },
}


items = [
    "[`{}`](./ingest/{}.md) [![image]({})](./ingest/{}.md) {}".format(
        dataset_name,
        dataset_name,
        details["thumbnail"],
        dataset_name,
        details["description"],
    )
    for dataset_name, details in list_of_datasets.items()
]



import os
from tqdm import tqdm, trange
from typing import List, Tuple, Dict
import numpy as np
import matplotlib.pyplot as plt
import random
import glob

from blueness import module
from blue_options import string
from blue_objects import file, path


from roofai import NAME, VERSION
from roofai.dataset.classes import RoofAIDataset, DatasetKind, MatrixKind, DatasetTarget
from roofai.logger import logger

NAME = module.name(__file__, NAME)


def ingest_from_dataset(
    input_dataset_path: str,
    output_dataset_path: str,
    counts: Dict[str, int],
    chip_overlap: float = 0.25,
    log: bool = False,
    verbose: bool = False,
    in_notebook: bool = False,
    target: DatasetTarget = DatasetTarget.TORCH,
    is_distributed: bool = False,
) -> bool:
    chip_height = target.chip_height
    chip_width = target.chip_width

    input_object_name = path.name(input_dataset_path)
    output_object_name = path.name(output_dataset_path)

    logger.info(
        "ingesting from {}{} -{}-{}x{}-@{:.0f}%-> {}:{}".format(
            input_object_name,
            " (distributed)" if is_distributed else "",
            " + ".join(
                ["{} X {:,d}".format(subset, count) for subset, count in counts.items()]
            ),
            chip_height,
            chip_width,
            chip_overlap * 100,
            target.name.lower(),
            output_object_name,
        )
    )

    original_counts = {class_name: count for class_name, count in counts.items()}
    if is_distributed:
        total_count = sum([count for count in counts.values()])
        counts = {
            class_name: total_count if class_name == "train" else 0
            for class_name, count in counts.items()
        }
        logger.info(f"☁️ distributed dataset, will ingest {total_count:,} chips first.")

    input_dataset = RoofAIDataset(input_dataset_path)
    output_dataset = RoofAIDataset(
        output_dataset_path,
        kind=(
            DatasetKind.CAMVID
            if target == DatasetTarget.TORCH
            else DatasetKind.SAGEMAKER
        ),
    ).create(log=log)

    output_dataset.classes = [class_name for class_name in input_dataset.classes]

    train_record_id_list = []
    for subset in tqdm(counts.keys()):
        record_id_list = []
        for matrix_kind in [MatrixKind.MASK, MatrixKind.IMAGE]:  # order is critical.
            chip_count = counts[subset]
            for record_id in input_dataset.subsets[subset]:
                input_matrix = input_dataset.get_matrix(
                    subset,
                    record_id,
                    matrix_kind,
                    log=log,
                )

                slice_count, slice_record_id_list = slice_matrix(
                    input_matrix=input_matrix,
                    kind=matrix_kind,
                    chip_height=chip_height,
                    chip_width=chip_width,
                    chip_overlap=chip_overlap,
                    max_chip_count=chip_count,
                    record_id_list=record_id_list,
                    output_path=output_dataset.subset_path(subset, matrix_kind),
                    target=target,
                    prefix=record_id,
                    log=log,
                    verbose=verbose,
                    class_count=len(output_dataset.classes),
                )

                chip_count -= slice_count

                record_id_list = list(set(record_id_list + slice_record_id_list))

                if chip_count <= 0:
                    break
                if log:
                    logger.info(f"remaining chip count: {chip_count:,}")

        if subset == "train":
            train_record_id_list = [record_id for record_id in record_id_list]

    if is_distributed:
        counts = {
            class_name: (
                int(original_count / total_count * len(train_record_id_list))
                if total_count
                else 0
            )
            for class_name, original_count in original_counts.items()
        }
        logger.info(
            "splitting {:,} chips to {}".format(
                len(train_record_id_list),
                " + ".join(
                    [f"{class_name}:{count:,}" for class_name, count in counts.items()]
                ),
            )
        )

        random.shuffle(train_record_id_list)

        global_record_index = counts["train"]
        for subset in tqdm(counts.keys()):
            if subset == "train":
                continue

            next_global_record_index = global_record_index + counts[subset]
            for record_index in trange(
                global_record_index,
                next_global_record_index,
            ):
                record_id = train_record_id_list[record_index]

                for matrix_kind in [MatrixKind.MASK, MatrixKind.IMAGE]:
                    path_pairs: Dict[str, str] = {
                        output_dataset.subset_path(
                            "train", matrix_kind
                        ): output_dataset.subset_path(subset, matrix_kind)
                    }

                    if matrix_kind == MatrixKind.MASK:
                        path_pairs.update(
                            {
                                f"{source_path}-colored": f"{destination_path}-colored"
                                for source_path, destination_path in path_pairs.items()
                            }
                        )

                    for source_path, destination_path in path_pairs.items():
                        for filename in glob.glob(
                            os.path.join(
                                source_path,
                                f"{record_id}*.*",
                            )
                        ):
                            if not file.move(
                                filename,
                                destination=os.path.join(
                                    destination_path,
                                    file.name_and_extension(filename),
                                ),
                            ):
                                return False

            global_record_index = next_global_record_index

    file.save_yaml(
        os.path.join(output_dataset_path, "metadata.yaml"),
        {
            "classes": output_dataset.classes,
            "kind": "CamVid" if target == DatasetTarget.TORCH else "SageMaker",
            "source": input_object_name if is_distributed else "AIRS",
            "ingested-by": f"{NAME}-{VERSION}",
            # SageMaker
            "bucket": "kamangir",
            "channel": (
                {
                    "label_map": f"s3://kamangir/bolt/{output_object_name}/label_map/train_label_map.json",
                    "train": f"s3://kamangir/bolt/{output_object_name}/train",
                    "train_annotation": f"s3://kamangir/bolt/{output_object_name}/train_annotation",
                    "validation": f"s3://kamangir/bolt/{output_object_name}/validation",
                    "validation_annotation": f"s3://kamangir/bolt/{output_object_name}/validation_annotation",
                }
                if target == DatasetTarget.SAGEMAKER
                else {}
            ),
            "num": counts,
            "prefix": f"bolt/{output_object_name}",
        },
        log=True,
    )

    RoofAIDataset(output_dataset_path).visualize(
        subset="train",
        index=0,
        in_notebook=in_notebook,
    )

    return True


def slice_matrix(
    input_matrix: np.ndarray,
    kind: MatrixKind,
    chip_height: int,
    chip_width: int,
    chip_overlap: float,
    max_chip_count: int,
    record_id_list: List[str],
    output_path: str,
    prefix: str,
    class_count: int,
    target: DatasetTarget = DatasetTarget.TORCH,
    log: bool = True,
    verbose: bool = False,
) -> Tuple[int, List[str]]:
    if log:
        logger.info(
            "slice_matrix[{}]: {} -{}X{}x{}-@{:.0f}%-> {} - {}{}".format(
                string.pretty_shape_of_matrix(input_matrix),
                kind,
                max_chip_count,
                chip_height,
                chip_width,
                chip_overlap * 100,
                output_path,
                prefix,
                (
                    ""
                    if kind == MatrixKind.MASK
                    else ": {} record_id(s): {}".format(
                        len(record_id_list),
                        ", ".join(record_id_list[:3] + ["..."]),
                    )
                ),
            )
        )

    record_id_list_output = []

    count = 0
    for y in range(
        0, input_matrix.shape[0] - chip_height, int(chip_overlap * chip_height)
    ):
        for x in range(
            0, input_matrix.shape[1] - chip_width, int(chip_overlap * chip_width)
        ):
            chip = input_matrix[
                y : y + chip_height,
                x : x + chip_width,
            ]

            record_id = f"{prefix}-{y:05d}-{x:05d}"

            # to ensure variety of labels in the pixel.
            # TODO: make it more elaborate.
            if (kind == MatrixKind.MASK and np.all(chip == 0)) or (
                kind == MatrixKind.IMAGE and (record_id not in record_id_list)
            ):
                continue
            record_id_list_output += [record_id]

            assert file.save_image(
                os.path.join(
                    output_path,
                    "{}.{}".format(
                        record_id,
                        (
                            "jpg"
                            if (kind == MatrixKind.IMAGE)
                            and (target == DatasetTarget.SAGEMAKER)
                            else "png"
                        ),
                    ),
                ),
                chip,
                log=verbose,
            )

            if kind == MatrixKind.MASK:
                assert file.save_image(
                    os.path.join(
                        path.parent(output_path),
                        f"{path.name(output_path)}-colored",
                        f"{record_id}.png",
                    ),
                    (
                        plt.cm.viridis(chip.astype(np.float32) / class_count) * 255
                    ).astype(np.uint8)[:, :, :3],
                    log=verbose,
                )

            count += 1
            if count >= max_chip_count:
                return count, record_id_list_output

    return count, record_id_list_output
